import unittest

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.test import TestCase, override_settings

from haystack import indexes
from haystack.exceptions import NotHandled, SearchFieldError
from haystack.utils import loading
from test_haystack.core.models import AnotherMockModel, MockModel

try:
    import pysolr
except ImportError:
    pysolr = False


class ConnectionHandlerTestCase(TestCase):
    def test_init(self):
        ch = loading.ConnectionHandler({})
        self.assertEqual(ch.connections_info, {})

        ch = loading.ConnectionHandler(
            {
                "default": {
                    "ENGINE": "haystack.backends.solr_backend.SolrEngine",
                    "URL": "http://localhost:9001/solr/test_default",
                }
            }
        )
        self.assertEqual(
            ch.connections_info,
            {
                "default": {
                    "ENGINE": "haystack.backends.solr_backend.SolrEngine",
                    "URL": "http://localhost:9001/solr/test_default",
                }
            },
        )

    @unittest.skipIf(pysolr is False, "pysolr required")
    def test_get_item(self):
        ch = loading.ConnectionHandler({})

        try:
            empty_engine = ch["default"]
            self.fail()
        except ImproperlyConfigured:
            pass

        ch = loading.ConnectionHandler(
            {
                "default": {
                    "ENGINE": "haystack.backends.solr_backend.SolrEngine",
                    "URL": "http://localhost:9001/solr/test_default",
                }
            }
        )
        solr_engine = ch["default"]
        backend_path, memory_address = (
            repr(solr_engine).strip("<>").split(" object at ")
        )
        self.assertEqual(backend_path, "haystack.backends.solr_backend.SolrEngine")

        solr_engine_2 = ch["default"]
        backend_path_2, memory_address_2 = (
            repr(solr_engine_2).strip("<>").split(" object at ")
        )
        self.assertEqual(backend_path_2, "haystack.backends.solr_backend.SolrEngine")
        # Ensure we're loading out of the memorized connection.
        self.assertEqual(memory_address_2, memory_address)

        try:
            empty_engine = ch["slave"]
            self.fail()
        except ImproperlyConfigured:
            pass

    def test_get_unified_index(self):
        ch = loading.ConnectionHandler(
            {"default": {"ENGINE": "haystack.backends.simple_backend.SimpleEngine"}}
        )
        ui = ch["default"].get_unified_index()
        klass, address = repr(ui).strip("<>").split(" object at ")
        self.assertEqual(str(klass), "haystack.utils.loading.UnifiedIndex")

        ui_2 = ch["default"].get_unified_index()
        klass_2, address_2 = repr(ui_2).strip("<>").split(" object at ")
        self.assertEqual(str(klass_2), "haystack.utils.loading.UnifiedIndex")
        self.assertEqual(address_2, address)


class ConnectionRouterTestCase(TestCase):
    @override_settings()
    def test_init(self):
        del settings.HAYSTACK_ROUTERS
        cr = loading.ConnectionRouter()
        self.assertEqual(
            [str(route.__class__) for route in cr.routers],
            ["<class 'haystack.routers.DefaultRouter'>"],
        )

    @override_settings(HAYSTACK_ROUTERS=["haystack.routers.DefaultRouter"])
    def test_router_override1(self):
        cr = loading.ConnectionRouter()
        self.assertEqual(
            [str(route.__class__) for route in cr.routers],
            ["<class 'haystack.routers.DefaultRouter'>"],
        )

    @override_settings(HAYSTACK_ROUTERS=[])
    def test_router_override2(self):
        cr = loading.ConnectionRouter()
        self.assertEqual(
            [str(route.__class__) for route in cr.routers],
            ["<class 'haystack.routers.DefaultRouter'>"],
        )

    @override_settings(
        HAYSTACK_ROUTERS=[
            "test_haystack.mocks.MockMasterSlaveRouter",
            "haystack.routers.DefaultRouter",
        ]
    )
    def test_router_override3(self):
        cr = loading.ConnectionRouter()
        self.assertEqual(
            [str(route.__class__) for route in cr.routers],
            [
                "<class 'test_haystack.mocks.MockMasterSlaveRouter'>",
                "<class 'haystack.routers.DefaultRouter'>",
            ],
        )

    @override_settings()
    def test_actions1(self):
        del settings.HAYSTACK_ROUTERS
        cr = loading.ConnectionRouter()
        self.assertEqual(cr.for_read(), "default")
        self.assertEqual(cr.for_write(), ["default"])

    @override_settings(
        HAYSTACK_ROUTERS=[
            "test_haystack.mocks.MockMasterSlaveRouter",
            "haystack.routers.DefaultRouter",
        ]
    )
    def test_actions2(self):
        cr = loading.ConnectionRouter()
        self.assertEqual(cr.for_read(), "slave")
        self.assertEqual(cr.for_write(), ["master", "default"])

    @override_settings(
        HAYSTACK_ROUTERS=[
            "test_haystack.mocks.MockPassthroughRouter",
            "test_haystack.mocks.MockMasterSlaveRouter",
            "haystack.routers.DefaultRouter",
        ]
    )
    def test_actions3(self):
        cr = loading.ConnectionRouter()
        # Demonstrate pass-through
        self.assertEqual(cr.for_read(), "slave")
        self.assertEqual(cr.for_write(), ["master", "default"])
        # Demonstrate that hinting can change routing.
        self.assertEqual(cr.for_read(pass_through=False), "pass")
        self.assertEqual(
            cr.for_write(pass_through=False), ["pass", "master", "default"]
        )

    @override_settings(
        HAYSTACK_ROUTERS=[
            "test_haystack.mocks.MockMultiRouter",
            "haystack.routers.DefaultRouter",
        ]
    )
    def test_actions4(self):
        cr = loading.ConnectionRouter()
        # Demonstrate that a router can return multiple backends in the "for_write" method
        self.assertEqual(cr.for_read(), "default")
        self.assertEqual(cr.for_write(), ["multi1", "multi2", "default"])


class MockNotAModel:
    pass


class FakeSearchIndex(indexes.BasicSearchIndex, indexes.Indexable):
    def update_object(self, instance, **kwargs):
        # Incorrect behavior but easy to test and all we care about is that we
        # make it here. We rely on the `SearchIndex` tests to ensure correct
        # behavior.
        return True

    def remove_object(self, instance, **kwargs):
        # Incorrect behavior but easy to test and all we care about is that we
        # make it here. We rely on the `SearchIndex` tests to ensure correct
        # behavior.
        return True

    def get_model(self):
        return MockModel


class InvalidSearchIndex(indexes.SearchIndex, indexes.Indexable):
    document = indexes.CharField(document=True)

    def get_model(self):
        return MockModel


class BasicMockModelSearchIndex(indexes.BasicSearchIndex, indexes.Indexable):
    def get_model(self):
        return MockModel


class BasicAnotherMockModelSearchIndex(indexes.BasicSearchIndex, indexes.Indexable):
    def get_model(self):
        return AnotherMockModel


class ValidSearchIndex(indexes.SearchIndex, indexes.Indexable):
    text = indexes.CharField(document=True)
    author = indexes.CharField(index_fieldname="name")
    title = indexes.CharField(indexed=False)

    def get_model(self):
        return MockModel


class AlternateValidSearchIndex(indexes.SearchIndex, indexes.Indexable):
    text = indexes.CharField(document=True)
    author = indexes.CharField(faceted=True)
    title = indexes.CharField(faceted=True)

    def get_model(self):
        return AnotherMockModel


class ExplicitFacetSearchIndex(indexes.SearchIndex, indexes.Indexable):
    text = indexes.CharField(document=True)
    author = indexes.CharField(faceted=True)
    title = indexes.CharField()
    title_facet = indexes.FacetCharField(facet_for="title")
    bare_facet = indexes.FacetCharField()

    def get_model(self):
        return MockModel


class MultiValueValidSearchIndex(indexes.SearchIndex, indexes.Indexable):
    text = indexes.CharField(document=True)
    author = indexes.MultiValueField(stored=False)
    title = indexes.CharField(indexed=False)

    def get_model(self):
        return MockModel


class UnifiedIndexTestCase(TestCase):
    def setUp(self):
        super().setUp()
        self.ui = loading.UnifiedIndex()
        self.ui.build([])

    def test_get_index(self):
        self.assertRaises(NotHandled, self.ui.get_index, MockModel)
        try:
            self.ui.get_index(MockModel)
        except NotHandled as e:
            self.assertTrue(MockModel.__name__ in str(e))

        self.ui.build(indexes=[BasicMockModelSearchIndex()])
        self.assertTrue(
            isinstance(self.ui.get_index(MockModel), indexes.BasicSearchIndex)
        )

    def test_get_indexed_models(self):
        self.assertEqual(self.ui.get_indexed_models(), [])

        self.ui.build(indexes=[ValidSearchIndex()])
        indexed_models = self.ui.get_indexed_models()
        self.assertEqual(len(indexed_models), 1)
        self.assertTrue(MockModel in indexed_models)

    def test_get_indexes(self):
        self.assertEqual(self.ui.get_indexes(), {})

        index = ValidSearchIndex()
        self.ui.build(indexes=[index])

        results = self.ui.get_indexes()
        self.assertEqual(len(results), 1)
        self.assertTrue(MockModel in results)
        self.assertEqual(results[MockModel], index)

    def test_all_searchfields(self):
        self.ui.build(indexes=[BasicMockModelSearchIndex()])
        fields = self.ui.all_searchfields()
        self.assertEqual(len(fields), 1)
        self.assertTrue("text" in fields)
        self.assertTrue(isinstance(fields["text"], indexes.CharField))
        self.assertEqual(fields["text"].document, True)
        self.assertEqual(fields["text"].use_template, True)

        self.ui.build(
            indexes=[BasicMockModelSearchIndex(), AlternateValidSearchIndex()]
        )
        fields = self.ui.all_searchfields()
        self.assertEqual(len(fields), 5)
        self.assertEqual(
            sorted(fields.keys()),
            ["author", "author_exact", "text", "title", "title_exact"],
        )
        self.assertTrue("text" in fields)
        self.assertTrue(isinstance(fields["text"], indexes.CharField))
        self.assertEqual(fields["text"].document, True)
        self.assertEqual(fields["text"].use_template, True)
        self.assertTrue("title" in fields)
        self.assertTrue(isinstance(fields["title"], indexes.CharField))
        self.assertEqual(fields["title"].document, False)
        self.assertEqual(fields["title"].use_template, False)
        self.assertEqual(fields["title"].faceted, True)
        self.assertEqual(fields["title"].indexed, True)
        self.assertTrue("author" in fields)
        self.assertTrue(isinstance(fields["author"], indexes.CharField))
        self.assertEqual(fields["author"].document, False)
        self.assertEqual(fields["author"].use_template, False)
        self.assertEqual(fields["author"].faceted, True)
        self.assertEqual(fields["author"].stored, True)
        self.assertEqual(fields["author"].index_fieldname, "author")

        self.ui.build(
            indexes=[AlternateValidSearchIndex(), MultiValueValidSearchIndex()]
        )
        fields = self.ui.all_searchfields()
        self.assertEqual(len(fields), 5)
        self.assertEqual(
            sorted(fields.keys()),
            ["author", "author_exact", "text", "title", "title_exact"],
        )
        self.assertTrue("text" in fields)
        self.assertTrue(isinstance(fields["text"], indexes.CharField))
        self.assertEqual(fields["text"].document, True)
        self.assertEqual(fields["text"].use_template, False)
        self.assertTrue("title" in fields)
        self.assertTrue(isinstance(fields["title"], indexes.CharField))
        self.assertEqual(fields["title"].document, False)
        self.assertEqual(fields["title"].use_template, False)
        self.assertEqual(fields["title"].faceted, True)
        self.assertEqual(fields["title"].indexed, True)
        self.assertTrue("author" in fields)
        self.assertTrue(isinstance(fields["author"], indexes.MultiValueField))
        self.assertEqual(fields["author"].document, False)
        self.assertEqual(fields["author"].use_template, False)
        self.assertEqual(fields["author"].stored, True)
        self.assertEqual(fields["author"].faceted, True)
        self.assertEqual(fields["author"].index_fieldname, "author")

        try:
            self.ui.build(indexes=[AlternateValidSearchIndex(), InvalidSearchIndex()])
            self.fail()
        except SearchFieldError:
            pass

    def test_get_index_fieldname(self):
        self.assertEqual(self.ui._fieldnames, {})

        self.ui.build(indexes=[ValidSearchIndex(), BasicAnotherMockModelSearchIndex()])
        self.ui.get_index_fieldname("text")
        self.assertEqual(
            self.ui._fieldnames, {"text": "text", "title": "title", "author": "name"}
        )
        self.assertEqual(self.ui.get_index_fieldname("text"), "text")
        self.assertEqual(self.ui.get_index_fieldname("author"), "name")
        self.assertEqual(self.ui.get_index_fieldname("title"), "title")

        # Reset the internal state to test the invalid case.
        self.ui.reset()
        self.assertEqual(self.ui._fieldnames, {})

        try:
            self.ui.build(indexes=[ValidSearchIndex(), AlternateValidSearchIndex()])
            self.fail()
        except SearchFieldError:
            pass

    def test_basic_get_facet_field_name(self):
        self.assertEqual(self.ui._facet_fieldnames, {})

        self.ui.build(
            indexes=[BasicMockModelSearchIndex(), AlternateValidSearchIndex()]
        )
        self.ui.get_facet_fieldname("text")
        self.assertEqual(
            self.ui._facet_fieldnames,
            {"title": "title_exact", "author": "author_exact"},
        )
        self.assertEqual(self.ui.get_index_fieldname("text"), "text")
        self.assertEqual(self.ui.get_index_fieldname("author"), "author")
        self.assertEqual(self.ui.get_index_fieldname("title"), "title")

        self.assertEqual(self.ui.get_facet_fieldname("text"), "text")
        self.assertEqual(self.ui.get_facet_fieldname("author"), "author_exact")
        self.assertEqual(self.ui.get_facet_fieldname("title"), "title_exact")

    def test_more_advanced_get_facet_field_name(self):
        self.assertEqual(self.ui._facet_fieldnames, {})

        self.ui.build(
            indexes=[BasicAnotherMockModelSearchIndex(), ExplicitFacetSearchIndex()]
        )
        self.ui.get_facet_fieldname("text")
        self.assertEqual(
            self.ui._facet_fieldnames,
            {
                "bare_facet": "bare_facet",
                "title": "title_facet",
                "author": "author_exact",
            },
        )
        self.assertEqual(self.ui.get_facet_fieldname("title"), "title_facet")
        self.assertEqual(self.ui.get_facet_fieldname("bare_facet"), "bare_facet")



import os
import unittest
from datetime import date, datetime, timedelta
from decimal import Decimal

from django.conf import settings
from django.test import TestCase
from django.test.utils import override_settings
from whoosh.analysis import SpaceSeparatedTokenizer, SubstitutionFilter
from whoosh.fields import BOOLEAN, DATETIME, KEYWORD, NUMERIC, TEXT
from whoosh.qparser import QueryParser

from haystack import connections, indexes, reset_search_queries
from haystack.exceptions import SearchBackendError, SkipDocument
from haystack.inputs import AutoQuery
from haystack.models import SearchResult
from haystack.query import SQ, SearchQuerySet
from haystack.utils.loading import UnifiedIndex

from ..core.models import AFourthMockModel, AnotherMockModel, MockModel
from ..mocks import MockSearchResult
from .testcases import WhooshTestCase


class WhooshMockSearchIndex(indexes.SearchIndex, indexes.Indexable):
    text = indexes.CharField(document=True, use_template=True)
    name = indexes.CharField(model_attr="author")
    pub_date = indexes.DateTimeField(model_attr="pub_date")
    name_analyzed = indexes.CharField(
        model_attr="author",
        analyzer=SpaceSeparatedTokenizer() | SubstitutionFilter(r"\d+", ""),
    )

    def get_model(self):
        return MockModel


class WhooshMockSearchIndexWithSkipDocument(WhooshMockSearchIndex):
    def prepare_text(self, obj):
        if obj.author == "daniel3":
            raise SkipDocument
        return obj.author


class WhooshAnotherMockSearchIndex(indexes.SearchIndex, indexes.Indexable):
    text = indexes.CharField(document=True)
    name = indexes.CharField(model_attr="author")
    pub_date = indexes.DateTimeField(model_attr="pub_date")

    def get_model(self):
        return AnotherMockModel

    def prepare_text(self, obj):
        return obj.author


class AllTypesWhooshMockSearchIndex(indexes.SearchIndex, indexes.Indexable):
    text = indexes.CharField(document=True, use_template=True)
    name = indexes.CharField(model_attr="author", indexed=False)
    pub_date = indexes.DateTimeField(model_attr="pub_date")
    sites = indexes.MultiValueField()
    seen_count = indexes.IntegerField(indexed=False)
    is_active = indexes.BooleanField(default=True)

    def get_model(self):
        return MockModel


class WhooshMaintainTypeMockSearchIndex(indexes.SearchIndex, indexes.Indexable):
    text = indexes.CharField(document=True)
    month = indexes.CharField(indexed=False)
    pub_date = indexes.DateTimeField(model_attr="pub_date")

    def get_model(self):
        return MockModel

    def prepare_text(self, obj):
        return "Indexed!\n%s" % obj.pk

    def prepare_month(self, obj):
        return "%02d" % obj.pub_date.month


class WhooshBoostMockSearchIndex(indexes.SearchIndex, indexes.Indexable):
    text = indexes.CharField(
        document=True,
        use_template=True,
        template_name="search/indexes/core/mockmodel_template.txt",
    )
    author = indexes.CharField(model_attr="author", weight=2.0)
    editor = indexes.CharField(model_attr="editor")
    pub_date = indexes.DateTimeField(model_attr="pub_date")

    def get_model(self):
        return AFourthMockModel

    def prepare(self, obj):
        data = super().prepare(obj)

        if obj.pk % 2 == 0:
            data["boost"] = 2.0

        return data


class WhooshAutocompleteMockModelSearchIndex(indexes.SearchIndex, indexes.Indexable):
    text = indexes.CharField(model_attr="foo", document=True)
    name = indexes.CharField(model_attr="author")
    pub_date = indexes.DateTimeField(model_attr="pub_date")
    text_auto = indexes.EdgeNgramField(model_attr="foo")
    name_auto = indexes.EdgeNgramField(model_attr="author")

    def get_model(self):
        return MockModel


@override_settings(USE_TZ=False)
class WhooshSearchBackendTestCase(WhooshTestCase):
    fixtures = ["bulk_data.json"]

    def setUp(self):
        super().setUp()

        self.old_ui = connections["whoosh"].get_unified_index()
        self.ui = UnifiedIndex()
        self.wmmi = WhooshMockSearchIndex()
        self.wmmidni = WhooshMockSearchIndexWithSkipDocument()
        self.wmtmmi = WhooshMaintainTypeMockSearchIndex()
        self.ui.build(indexes=[self.wmmi])
        self.sb = connections["whoosh"].get_backend()
        connections["whoosh"]._index = self.ui

        self.sb.setup()
        self.raw_whoosh = self.sb.index
        self.parser = QueryParser(self.sb.content_field_name, schema=self.sb.schema)
        self.sb.delete_index()

        self.sample_objs = MockModel.objects.all()

    def tearDown(self):
        connections["whoosh"]._index = self.old_ui
        super().tearDown()

    def whoosh_search(self, query):
        self.raw_whoosh = self.raw_whoosh.refresh()
        searcher = self.raw_whoosh.searcher()
        return searcher.search(self.parser.parse(query), limit=1000)

    def test_non_silent(self):
        bad_sb = connections["whoosh"].backend(
            "bad", PATH="/tmp/bad_whoosh", SILENTLY_FAIL=False
        )
        bad_sb.use_file_storage = False
        bad_sb.storage = "omg.wtf.bbq"

        try:
            bad_sb.update(self.wmmi, self.sample_objs)
            self.fail()
        except:
            pass

        try:
            bad_sb.remove("core.mockmodel.1")
            self.fail()
        except:
            pass

        try:
            bad_sb.clear()
            self.fail()
        except:
            pass

        try:
            bad_sb.search("foo")
            self.fail()
        except:
            pass

    def test_update(self):
        self.sb.update(self.wmmi, self.sample_objs)

        # Check what Whoosh thinks is there.
        self.assertEqual(len(self.whoosh_search("*")), 23)
        self.assertEqual(
            [doc.fields()["id"] for doc in self.whoosh_search("*")],
            ["core.mockmodel.%s" % i for i in range(1, 24)],
        )

    def test_update_with_SkipDocument_raised(self):
        self.sb.update(self.wmmidni, self.sample_objs)

        # Check what Whoosh thinks is there.
        res = self.whoosh_search("*")
        self.assertEqual(len(res), 14)
        ids = [1, 2, 5, 6, 7, 8, 9, 11, 12, 14, 15, 18, 20, 21]
        self.assertListEqual(
            [doc.fields()["id"] for doc in res], ["core.mockmodel.%s" % i for i in ids]
        )

    def test_remove(self):
        self.sb.update(self.wmmi, self.sample_objs)
        self.assertEqual(self.sb.index.doc_count(), 23)

        self.sb.remove(self.sample_objs[0])
        self.assertEqual(self.sb.index.doc_count(), 22)

    def test_clear(self):
        self.sb.update(self.wmmi, self.sample_objs)
        self.assertEqual(self.sb.index.doc_count(), 23)

        self.sb.clear()
        self.assertEqual(self.sb.index.doc_count(), 0)

        self.sb.update(self.wmmi, self.sample_objs)
        self.assertEqual(self.sb.index.doc_count(), 23)

        self.sb.clear([AnotherMockModel])
        self.assertEqual(self.sb.index.doc_count(), 23)

        self.sb.clear([MockModel])
        self.assertEqual(self.sb.index.doc_count(), 0)

        self.sb.index.refresh()
        self.sb.update(self.wmmi, self.sample_objs)
        self.assertEqual(self.sb.index.doc_count(), 23)

        self.sb.clear([AnotherMockModel, MockModel])
        self.assertEqual(self.raw_whoosh.doc_count(), 0)

    def test_search(self):
        self.sb.update(self.wmmi, self.sample_objs)
        self.assertEqual(len(self.whoosh_search("*")), 23)

        # No query string should always yield zero results.
        self.assertEqual(self.sb.search(""), {"hits": 0, "results": []})

        # A one letter query string gets nabbed by a stopwords filter. Should
        # always yield zero results.
        self.assertEqual(self.sb.search("a"), {"hits": 0, "results": []})

        # Possible AttributeError?
        # self.assertEqual(self.sb.search(u'a b'), {'hits': 0, 'results': [], 'spelling_suggestion': '', 'facets': {}})

        self.assertEqual(self.sb.search("*")["hits"], 23)
        self.assertEqual(
            [result.pk for result in self.sb.search("*")["results"]],
            ["%s" % i for i in range(1, 24)],
        )

        self.assertEqual(self.sb.search("Indexe")["hits"], 23)
        self.assertEqual(self.sb.search("Indexe")["spelling_suggestion"], "indexed")

        self.assertEqual(
            self.sb.search("", facets=["name"]), {"hits": 0, "results": []}
        )
        results = self.sb.search("index*", facets=["name"])
        self.assertEqual(results["hits"], 23)
        self.assertEqual(results["facets"]["dates"], {})
        self.assertEqual(results["facets"]["queries"], {})
        self.assertEqual(
            results["facets"]["fields"]["name"],
            [("daniel3", 9), ("daniel1", 7), ("daniel2", 7)],
        )

        self.assertEqual(
            self.sb.search(
                "",
                date_facets={
                    "pub_date": {
                        "start_date": date(2007, 2, 26),
                        "end_date": date(2008, 2, 26),
                        "gap_by": "month",
                    }
                },
            ),
            {"hits": 0, "results": []},
        )
        results = self.sb.search(
            "Index*",
            date_facets={
                "pub_date": {
                    "start_date": date(2007, 2, 26),
                    "end_date": date(2008, 2, 26),
                    "gap_by": "month",
                }
            },
        )
        results = self.sb.search(
            "index*",
            date_facets={
                "pub_date": {
                    "start_date": date(2007, 2, 26),
                    "end_date": date(2008, 2, 26),
                    "gap_by": "month",
                }
            },
        )
        self.assertEqual(results["hits"], 23)
        self.assertEqual(results["facets"]["fields"], {})
        self.assertEqual(results["facets"]["queries"], {})
        self.assertEqual(results["facets"]["dates"]["pub_date"], [(None, 23)])

        results = self.sb.search(
            "index*",
            date_facets={
                "pub_date": {
                    "start_date": date(2009, 3, 26),
                    "end_date": date(2010, 2, 26),
                    "gap_by": "month",
                    "gap_amount": 2,
                }
            },
        )
        self.assertEqual(results["hits"], 23)
        self.assertEqual(
            results["facets"]["dates"]["pub_date"],
            [
                ((datetime(2009, 5, 26, 0, 0), datetime(2009, 7, 26, 0, 0)), 23),
            ],
        )

        results = self.sb.search(
            "index*",
            date_facets={
                "pub_date": {
                    "start_date": date(2009, 7, 1),
                    "end_date": date(2009, 8, 1),
                    "gap_by": "day",
                    "gap_amount": 1,
                }
            },
        )
        self.assertEqual(results["hits"], 23)
        self.assertEqual(
            results["facets"]["dates"]["pub_date"],
            [
                ((datetime(2009, 7, 17, 0, 0), datetime(2009, 7, 18, 0, 0)), 21),
                (None, 2),
            ],
        )

        results = self.sb.search(
            "index*",
            date_facets={
                "pub_date": {
                    "start_date": datetime(2009, 6, 1),
                    "end_date": datetime(2009, 8, 1),
                    "gap_by": "hour",
                }
            },
        )
        self.assertEqual(results["hits"], 23)
        self.assertEqual(
            results["facets"]["dates"]["pub_date"],
            [
                ((datetime(2009, 6, 18, 6, 0), datetime(2009, 6, 18, 7, 0)), 1),
                ((datetime(2009, 6, 18, 8, 0), datetime(2009, 6, 18, 9, 0)), 1),
                ((datetime(2009, 7, 17, 0, 0), datetime(2009, 7, 17, 1, 0)), 1),
                ((datetime(2009, 7, 17, 1, 0), datetime(2009, 7, 17, 2, 0)), 1),
                ((datetime(2009, 7, 17, 2, 0), datetime(2009, 7, 17, 3, 0)), 1),
                ((datetime(2009, 7, 17, 3, 0), datetime(2009, 7, 17, 4, 0)), 1),
                ((datetime(2009, 7, 17, 4, 0), datetime(2009, 7, 17, 5, 0)), 1),
                ((datetime(2009, 7, 17, 5, 0), datetime(2009, 7, 17, 6, 0)), 1),
                ((datetime(2009, 7, 17, 6, 0), datetime(2009, 7, 17, 7, 0)), 1),
                ((datetime(2009, 7, 17, 7, 0), datetime(2009, 7, 17, 8, 0)), 1),
                ((datetime(2009, 7, 17, 8, 0), datetime(2009, 7, 17, 9, 0)), 1),
                ((datetime(2009, 7, 17, 9, 0), datetime(2009, 7, 17, 10, 0)), 1),
                ((datetime(2009, 7, 17, 10, 0), datetime(2009, 7, 17, 11, 0)), 1),
                ((datetime(2009, 7, 17, 11, 0), datetime(2009, 7, 17, 12, 0)), 1),
                ((datetime(2009, 7, 17, 12, 0), datetime(2009, 7, 17, 13, 0)), 1),
                ((datetime(2009, 7, 17, 13, 0), datetime(2009, 7, 17, 14, 0)), 1),
                ((datetime(2009, 7, 17, 14, 0), datetime(2009, 7, 17, 15, 0)), 1),
                ((datetime(2009, 7, 17, 15, 0), datetime(2009, 7, 17, 16, 0)), 1),
                ((datetime(2009, 7, 17, 16, 0), datetime(2009, 7, 17, 17, 0)), 1),
                ((datetime(2009, 7, 17, 17, 0), datetime(2009, 7, 17, 18, 0)), 1),
                ((datetime(2009, 7, 17, 18, 0), datetime(2009, 7, 17, 19, 0)), 1),
                ((datetime(2009, 7, 17, 19, 0), datetime(2009, 7, 17, 20, 0)), 1),
                ((datetime(2009, 7, 17, 20, 0), datetime(2009, 7, 17, 21, 0)), 1),
            ],
        )

        self.assertEqual(
            self.sb.search("", query_facets={"name": "[* TO e]"}),
            {"hits": 0, "results": []},
        )
        results = self.sb.search("Index*", query_facets={"name": "[* TO e]"})
        results = self.sb.search("index*", query_facets={"name": "[* TO e]"})
        self.assertEqual(results["hits"], 23)
        self.assertEqual(results["facets"], {})

        self.assertEqual(
            self.sb.search("", narrow_queries=set(["name:daniel1"])),
            {"hits": 0, "results": []},
        )
        results = self.sb.search("Index*", narrow_queries=set(["name:daniel1"]))
        self.assertEqual(results["hits"], 7)

        # Ensure that swapping the ``result_class`` works.
        self.assertTrue(
            isinstance(
                self.sb.search("Index*", result_class=MockSearchResult)["results"][0],
                MockSearchResult,
            )
        )

        # Check the use of ``limit_to_registered_models``.
        self.assertEqual(
            self.sb.search("", limit_to_registered_models=False),
            {"hits": 0, "results": []},
        )
        self.assertEqual(
            self.sb.search("*", limit_to_registered_models=False)["hits"], 23
        )
        self.assertEqual(
            [
                result.pk
                for result in self.sb.search("*", limit_to_registered_models=False)[
                    "results"
                ]
            ],
            ["%s" % i for i in range(1, 24)],
        )

        # Stow.
        old_limit_to_registered_models = getattr(
            settings, "HAYSTACK_LIMIT_TO_REGISTERED_MODELS", True
        )
        settings.HAYSTACK_LIMIT_TO_REGISTERED_MODELS = False

        self.assertEqual(self.sb.search(""), {"hits": 0, "results": []})
        self.assertEqual(self.sb.search("*")["hits"], 23)
        self.assertEqual(
            [result.pk for result in self.sb.search("*")["results"]],
            ["%s" % i for i in range(1, 24)],
        )

        # Restore.
        settings.HAYSTACK_LIMIT_TO_REGISTERED_MODELS = old_limit_to_registered_models

    def test_highlight(self):
        self.sb.update(self.wmmi, self.sample_objs)
        self.assertEqual(len(self.whoosh_search("*")), 23)

        self.assertEqual(self.sb.search("", highlight=True), {"hits": 0, "results": []})
        self.assertEqual(self.sb.search("index*", highlight=True)["hits"], 23)

        query = self.sb.search("Index*", highlight=True)["results"]
        result = [result.highlighted["text"][0] for result in query]

        self.assertEqual(result, ["<em>Indexed</em>!\n%d" % i for i in range(1, 24)])

    def test_search_all_models(self):
        wamsi = WhooshAnotherMockSearchIndex()
        self.ui.build(indexes=[self.wmmi, wamsi])

        self.sb.update(self.wmmi, self.sample_objs)
        self.sb.update(wamsi, AnotherMockModel.objects.all())

        self.assertEqual(len(self.whoosh_search("*")), 25)

        self.ui.build(indexes=[self.wmmi])

    def test_more_like_this(self):
        self.sb.update(self.wmmi, self.sample_objs)
        self.assertEqual(len(self.whoosh_search("*")), 23)

        # Now supported by Whoosh (as of 1.8.4). See the ``LiveWhooshMoreLikeThisTestCase``.
        self.assertEqual(self.sb.more_like_this(self.sample_objs[0])["hits"], 22)

        # Make sure that swapping the ``result_class`` doesn't blow up.
        try:
            self.sb.more_like_this(self.sample_objs[0], result_class=MockSearchResult)
        except:
            self.fail()

    def test_delete_index(self):
        self.sb.update(self.wmmi, self.sample_objs)
        self.assertTrue(self.sb.index.doc_count() > 0)

        self.sb.delete_index()
        self.assertEqual(self.sb.index.doc_count(), 0)

    def test_order_by(self):
        self.sb.update(self.wmmi, self.sample_objs)

        results = self.sb.search("*", sort_by=["pub_date"])
        self.assertEqual(
            [result.pk for result in results["results"]],
            [
                "1",
                "3",
                "2",
                "4",
                "5",
                "6",
                "7",
                "8",
                "9",
                "10",
                "11",
                "12",
                "13",
                "14",
                "15",
                "16",
                "17",
                "18",
                "19",
                "20",
                "21",
                "22",
                "23",
            ],
        )

        results = self.sb.search("*", sort_by=["-pub_date"])
        self.assertEqual(
            [result.pk for result in results["results"]],
            [
                "23",
                "22",
                "21",
                "20",
                "19",
                "18",
                "17",
                "16",
                "15",
                "14",
                "13",
                "12",
                "11",
                "10",
                "9",
                "8",
                "7",
                "6",
                "5",
                "4",
                "2",
                "3",
                "1",
            ],
        )

        results = self.sb.search("*", sort_by=["id"])
        self.assertEqual(
            [result.pk for result in results["results"]],
            [
                "1",
                "10",
                "11",
                "12",
                "13",
                "14",
                "15",
                "16",
                "17",
                "18",
                "19",
                "2",
                "20",
                "21",
                "22",
                "23",
                "3",
                "4",
                "5",
                "6",
                "7",
                "8",
                "9",
            ],
        )

        results = self.sb.search("*", sort_by=["-id"])
        self.assertEqual(
            [result.pk for result in results["results"]],
            [
                "9",
                "8",
                "7",
                "6",
                "5",
                "4",
                "3",
                "23",
                "22",
                "21",
                "20",
                "2",
                "19",
                "18",
                "17",
                "16",
                "15",
                "14",
                "13",
                "12",
                "11",
                "10",
                "1",
            ],
        )

        results = self.sb.search("*", sort_by=["-pub_date", "-id"])
        self.assertEqual(
            [result.pk for result in results["results"]],
            [
                "23",
                "22",
                "21",
                "20",
                "19",
                "18",
                "17",
                "16",
                "15",
                "14",
                "13",
                "12",
                "11",
                "10",
                "9",
                "8",
                "7",
                "6",
                "5",
                "4",
                "2",
                "3",
                "1",
            ],
        )

        self.assertRaises(
            SearchBackendError, self.sb.search, "*", sort_by=["-pub_date", "id"]
        )

    def test__from_python(self):
        self.assertEqual(self.sb._from_python("abc"), "abc")
        self.assertEqual(self.sb._from_python(1), 1)
        self.assertEqual(self.sb._from_python(2653), 2653)
        self.assertEqual(self.sb._from_python(25.5), 25.5)
        self.assertEqual(self.sb._from_python([1, 2, 3]), "1,2,3")
        self.assertTrue("a': 1" in self.sb._from_python({"a": 1, "c": 3, "b": 2}))
        self.assertEqual(
            self.sb._from_python(datetime(2009, 5, 9, 16, 14)),
            datetime(2009, 5, 9, 16, 14),
        )
        self.assertEqual(
            self.sb._from_python(datetime(2009, 5, 9, 0, 0)), datetime(2009, 5, 9, 0, 0)
        )
        self.assertEqual(
            self.sb._from_python(datetime(1899, 5, 18, 0, 0)),
            datetime(1899, 5, 18, 0, 0),
        )
        self.assertEqual(
            self.sb._from_python(datetime(2009, 5, 18, 1, 16, 30, 250)),
            datetime(2009, 5, 18, 1, 16, 30, 250),
        )

    def test__to_python(self):
        self.assertEqual(self.sb._to_python("abc"), "abc")
        self.assertEqual(self.sb._to_python("1"), 1)
        self.assertEqual(self.sb._to_python("2653"), 2653)
        self.assertEqual(self.sb._to_python("25.5"), 25.5)
        self.assertEqual(self.sb._to_python("[1, 2, 3]"), [1, 2, 3])
        self.assertEqual(
            self.sb._to_python('{"a": 1, "b": 2, "c": 3}'), {"a": 1, "c": 3, "b": 2}
        )
        self.assertEqual(
            self.sb._to_python("2009-05-09T16:14:00"), datetime(2009, 5, 9, 16, 14)
        )
        self.assertEqual(
            self.sb._to_python("2009-05-09T00:00:00"), datetime(2009, 5, 9, 0, 0)
        )
        self.assertEqual(self.sb._to_python(None), None)

    def test_range_queries(self):
        self.sb.update(self.wmmi, self.sample_objs)

        self.assertEqual(len(self.whoosh_search("[d TO]")), 23)
        self.assertEqual(len(self.whoosh_search("name:[d TO]")), 23)
        self.assertEqual(len(self.whoosh_search("Ind* AND name:[d to]")), 23)
        self.assertEqual(len(self.whoosh_search("Ind* AND name:[to c]")), 0)

    def test_date_queries(self):
        self.sb.update(self.wmmi, self.sample_objs)

        self.assertEqual(len(self.whoosh_search("pub_date:20090717003000")), 1)
        self.assertEqual(len(self.whoosh_search("pub_date:20090717000000")), 0)
        self.assertEqual(
            len(self.whoosh_search("Ind* AND pub_date:[to 20090717003000]")), 3
        )

    def test_escaped_characters_queries(self):
        self.sb.update(self.wmmi, self.sample_objs)

        self.assertEqual(len(self.whoosh_search("Indexed\!")), 23)
        self.assertEqual(len(self.whoosh_search("http\:\/\/www\.example\.com")), 0)

    def test_build_schema(self):
        ui = UnifiedIndex()
        ui.build(indexes=[AllTypesWhooshMockSearchIndex()])

        (content_field_name, schema) = self.sb.build_schema(ui.all_searchfields())
        self.assertEqual(content_field_name, "text")

        schema_names = set(schema.names())
        required_schema = {
            "django_ct",
            "django_id",
            "id",
            "is_active",
            "name",
            "pub_date",
            "seen_count",
            "sites",
            "text",
        }
        self.assertTrue(required_schema.issubset(schema_names))

        self.assertIsInstance(schema._fields["text"], TEXT)
        self.assertIsInstance(schema._fields["pub_date"], DATETIME)
        self.assertIsInstance(schema._fields["seen_count"], NUMERIC)
        self.assertIsInstance(schema._fields["sites"], KEYWORD)
        self.assertIsInstance(schema._fields["is_active"], BOOLEAN)

    def test_verify_type(self):
        old_ui = connections["whoosh"].get_unified_index()
        ui = UnifiedIndex()
        wmtmmi = WhooshMaintainTypeMockSearchIndex()
        ui.build(indexes=[wmtmmi])
        connections["whoosh"]._index = ui
        sb = connections["whoosh"].get_backend()
        sb.setup()
        sb.update(wmtmmi, self.sample_objs)

        self.assertEqual(sb.search("*")["hits"], 23)
        self.assertEqual(
            [result.month for result in sb.search("*")["results"]],
            [
                "06",
                "07",
                "06",
                "07",
                "07",
                "07",
                "07",
                "07",
                "07",
                "07",
                "07",
                "07",
                "07",
                "07",
                "07",
                "07",
                "07",
                "07",
                "07",
                "07",
                "07",
                "07",
                "07",
            ],
        )
        connections["whoosh"]._index = old_ui

    @unittest.skipIf(
        settings.HAYSTACK_CONNECTIONS["whoosh"].get("STORAGE") != "file",
        "testing writability requires Whoosh to use STORAGE=file",
    )
    def test_writable(self):
        if not os.path.exists(settings.HAYSTACK_CONNECTIONS["whoosh"]["PATH"]):
            os.makedirs(settings.HAYSTACK_CONNECTIONS["whoosh"]["PATH"])

        os.chmod(settings.HAYSTACK_CONNECTIONS["whoosh"]["PATH"], 0o400)

        try:
            self.sb.setup()
            self.fail()
        except IOError:
            # Yay. We failed
            pass

        os.chmod(settings.HAYSTACK_CONNECTIONS["whoosh"]["PATH"], 0o755)

    def test_slicing(self):
        self.sb.update(self.wmmi, self.sample_objs)

        page_1 = self.sb.search("*", start_offset=0, end_offset=20)
        page_2 = self.sb.search("*", start_offset=20, end_offset=30)
        self.assertEqual(len(page_1["results"]), 20)
        self.assertEqual(
            [result.pk for result in page_1["results"]],
            ["%s" % i for i in range(1, 21)],
        )
        self.assertEqual(len(page_2["results"]), 3)
        self.assertEqual(
            [result.pk for result in page_2["results"]], ["21", "22", "23"]
        )

        # This used to throw an error.
        page_0 = self.sb.search("*", start_offset=0, end_offset=0)
        self.assertEqual(len(page_0["results"]), 1)

    @unittest.expectedFailure
    def test_scoring(self):
        self.sb.update(self.wmmi, self.sample_objs)

        page_1 = self.sb.search("index", start_offset=0, end_offset=20)
        page_2 = self.sb.search("index", start_offset=20, end_offset=30)
        self.assertEqual(len(page_1["results"]), 20)
        self.assertEqual(
            ["%0.2f" % result.score for result in page_1["results"]],
            [
                "0.51",
                "0.51",
                "0.51",
                "0.51",
                "0.51",
                "0.51",
                "0.51",
                "0.51",
                "0.51",
                "0.40",
                "0.40",
                "0.40",
                "0.40",
                "0.40",
                "0.40",
                "0.40",
                "0.40",
                "0.40",
                "0.40",
                "0.40",
            ],
        )
        self.assertEqual(len(page_2["results"]), 3)
        self.assertEqual(
            ["%0.2f" % result.score for result in page_2["results"]],
            ["0.40", "0.40", "0.40"],
        )

    def test_analyzed_fields(self):
        self.sb.update(self.wmmi, self.sample_objs)
        results = self.whoosh_search("name_analyzed:1234daniel5678")
        self.assertEqual(len(results), 23)


class WhooshBoostBackendTestCase(WhooshTestCase):
    def setUp(self):
        super().setUp()

        self.old_ui = connections["whoosh"].get_unified_index()
        self.ui = UnifiedIndex()
        self.wmmi = WhooshBoostMockSearchIndex()
        self.ui.build(indexes=[self.wmmi])
        self.sb = connections["whoosh"].get_backend()
        connections["whoosh"]._index = self.ui

        self.sb.setup()
        self.raw_whoosh = self.sb.index
        self.parser = QueryParser(self.sb.content_field_name, schema=self.sb.schema)
        self.sb.delete_index()
        self.sample_objs = []

        for i in range(1, 5):
            mock = AFourthMockModel()
            mock.id = i

            if i % 2:
                mock.author = "daniel"
                mock.editor = "david"
            else:
                mock.author = "david"
                mock.editor = "daniel"

            mock.pub_date = date(2009, 2, 25) - timedelta(days=i)
            self.sample_objs.append(mock)

    def tearDown(self):
        connections["whoosh"]._index = self.ui
        super().tearDown()

    @unittest.expectedFailure
    def test_boost(self):
        self.sb.update(self.wmmi, self.sample_objs)
        self.raw_whoosh = self.raw_whoosh.refresh()
        searcher = self.raw_whoosh.searcher()
        self.assertEqual(len(searcher.search(self.parser.parse("*"), limit=1000)), 2)

        results = SearchQuerySet("whoosh").filter(
            SQ(author="daniel") | SQ(editor="daniel")
        )

        self.assertEqual(
            [result.id for result in results],
            ["core.afourthmockmodel.1", "core.afourthmockmodel.3"],
        )
        self.assertEqual(results[0].boost, 1.1)


class LiveWhooshSearchQueryTestCase(WhooshTestCase):
    def setUp(self):
        super().setUp()

        # Stow.
        self.old_ui = connections["whoosh"].get_unified_index()
        self.ui = UnifiedIndex()
        self.wmmi = WhooshMockSearchIndex()
        self.wmtmmi = WhooshMaintainTypeMockSearchIndex()
        self.ui.build(indexes=[self.wmmi])
        self.sb = connections["whoosh"].get_backend()
        connections["whoosh"]._index = self.ui

        self.sb.setup()
        self.raw_whoosh = self.sb.index
        self.parser = QueryParser(self.sb.content_field_name, schema=self.sb.schema)
        self.sb.delete_index()

        self.sample_objs = []

        for i in range(1, 4):
            mock = MockModel()
            mock.id = i
            mock.author = "daniel%s" % i
            mock.pub_date = date(2009, 2, 25) - timedelta(days=i)
            self.sample_objs.append(mock)

        self.sq = connections["whoosh"].get_query()

    def tearDown(self):
        connections["whoosh"]._index = self.old_ui
        super().tearDown()

    def test_get_spelling(self):
        self.sb.update(self.wmmi, self.sample_objs)

        self.sq.add_filter(SQ(content="Indexe"))
        self.assertEqual(self.sq.get_spelling_suggestion(), "indexed")

    def test_log_query(self):
        from django.conf import settings

        reset_search_queries()
        self.assertEqual(len(connections["whoosh"].queries), 0)

        # Stow.

        with self.settings(DEBUG=False):
            len(self.sq.get_results())
            self.assertEqual(len(connections["whoosh"].queries), 0)

        with self.settings(DEBUG=True):
            # Redefine it to clear out the cached results.
            self.sq = connections["whoosh"].get_query()
            self.sq.add_filter(SQ(name="bar"))
            len(self.sq.get_results())
            self.assertEqual(len(connections["whoosh"].queries), 1)
            self.assertEqual(
                connections["whoosh"].queries[0]["query_string"], "name:(bar)"
            )

            # And again, for good measure.
            self.sq = connections["whoosh"].get_query()
            self.sq.add_filter(SQ(name="baz"))
            self.sq.add_filter(SQ(text="foo"))
            len(self.sq.get_results())
            self.assertEqual(len(connections["whoosh"].queries), 2)
            self.assertEqual(
                connections["whoosh"].queries[0]["query_string"], "name:(bar)"
            )
            self.assertEqual(
                connections["whoosh"].queries[1]["query_string"],
                "(name:(baz) AND text:(foo))",
            )


@override_settings(DEBUG=True)
class LiveWhooshSearchQuerySetTestCase(WhooshTestCase):
    def setUp(self):
        super().setUp()

        # Stow.
        self.old_ui = connections["whoosh"].get_unified_index()
        self.ui = UnifiedIndex()
        self.wmmi = WhooshMockSearchIndex()
        self.ui.build(indexes=[self.wmmi])
        self.sb = connections["whoosh"].get_backend()
        connections["whoosh"]._index = self.ui

        self.sb.setup()
        self.raw_whoosh = self.sb.index
        self.parser = QueryParser(self.sb.content_field_name, schema=self.sb.schema)
        self.sb.delete_index()

        self.sample_objs = []

        for i in range(1, 4):
            mock = MockModel()
            mock.id = i
            mock.author = "daniel%s" % i
            mock.pub_date = date(2009, 2, 25) - timedelta(days=i)
            self.sample_objs.append(mock)

        self.sq = connections["whoosh"].get_query()
        self.sqs = SearchQuerySet("whoosh")

    def tearDown(self):
        connections["whoosh"]._index = self.old_ui
        super().tearDown()

    def test_various_searchquerysets(self):
        self.sb.update(self.wmmi, self.sample_objs)

        sqs = self.sqs.filter(content="Index")
        self.assertEqual(sqs.query.build_query(), "(Index)")
        self.assertEqual(len(sqs), 3)

        sqs = self.sqs.auto_query("Indexed!")
        self.assertEqual(sqs.query.build_query(), "('Indexed!')")
        self.assertEqual(len(sqs), 3)

        sqs = self.sqs.auto_query("Indexed!").filter(pub_date__lte=date(2009, 8, 31))
        self.assertEqual(
            sqs.query.build_query(), "(('Indexed!') AND pub_date:([to 20090831000000]))"
        )
        self.assertEqual(len(sqs), 3)

        sqs = self.sqs.auto_query("Indexed!").filter(pub_date__lte=date(2009, 2, 23))
        self.assertEqual(
            sqs.query.build_query(), "(('Indexed!') AND pub_date:([to 20090223000000]))"
        )
        self.assertEqual(len(sqs), 2)

        sqs = (
            self.sqs.auto_query("Indexed!")
            .filter(pub_date__lte=date(2009, 2, 25))
            .filter(django_id__in=[1, 2])
            .exclude(name="daniel1")
        )
        self.assertEqual(
            sqs.query.build_query(),
            "(('Indexed!') AND pub_date:([to 20090225000000]) AND django_id:(1 OR 2) AND NOT (name:(daniel1)))",
        )
        self.assertEqual(len(sqs), 1)

        sqs = self.sqs.auto_query("re-inker")
        self.assertEqual(sqs.query.build_query(), "('re-inker')")
        self.assertEqual(len(sqs), 0)

        sqs = self.sqs.auto_query("0.7 wire")
        self.assertEqual(sqs.query.build_query(), "('0.7' wire)")
        self.assertEqual(len(sqs), 0)

        sqs = self.sqs.auto_query("daler-rowney pearlescent 'bell bronze'")
        self.assertEqual(
            sqs.query.build_query(), "('daler-rowney' pearlescent 'bell bronze')"
        )
        self.assertEqual(len(sqs), 0)

        sqs = self.sqs.models(MockModel)
        self.assertEqual(sqs.query.build_query(), "*")
        self.assertEqual(len(sqs), 3)

    def test_all_regression(self):
        sqs = SearchQuerySet("whoosh")
        self.assertEqual([result.pk for result in sqs], [])

        self.sb.update(self.wmmi, self.sample_objs)
        self.assertTrue(self.sb.index.doc_count() > 0)

        sqs = SearchQuerySet("whoosh")
        self.assertEqual(len(sqs), 3)
        self.assertEqual(sorted([result.pk for result in sqs]), ["1", "2", "3"])

        try:
            sqs = repr(SearchQuerySet("whoosh"))
        except:
            self.fail()

    def test_regression_space_query(self):
        self.sb.update(self.wmmi, self.sample_objs)
        self.assertTrue(self.sb.index.doc_count() > 0)

        sqs = SearchQuerySet("whoosh").auto_query(" ")
        self.assertEqual(len(sqs), 3)
        sqs = SearchQuerySet("whoosh").filter(content=" ")
        self.assertEqual(len(sqs), 0)

    def test_iter(self):
        self.sb.update(self.wmmi, self.sample_objs)

        reset_search_queries()
        self.assertEqual(len(connections["whoosh"].queries), 0)
        sqs = self.sqs.auto_query("Indexed!")
        results = [int(result.pk) for result in iter(sqs)]
        self.assertEqual(sorted(results), [1, 2, 3])
        self.assertEqual(len(connections["whoosh"].queries), 1)

    def test_slice(self):
        self.sb.update(self.wmmi, self.sample_objs)

        reset_search_queries()
        self.assertEqual(len(connections["whoosh"].queries), 0)
        results = self.sqs.auto_query("Indexed!")
        self.assertEqual(sorted([int(result.pk) for result in results[1:3]]), [1, 2])
        self.assertEqual(len(connections["whoosh"].queries), 1)

        reset_search_queries()
        self.assertEqual(len(connections["whoosh"].queries), 0)
        results = self.sqs.auto_query("Indexed!")
        self.assertEqual(int(results[0].pk), 1)
        self.assertEqual(len(connections["whoosh"].queries), 1)

    def test_values_slicing(self):
        self.sb.update(self.wmmi, self.sample_objs)

        reset_search_queries()
        self.assertEqual(len(connections["whoosh"].queries), 0)

        # TODO: this would be a good candidate for refactoring into a TestCase subclass shared across backends

        # The values will come back as strings because Hasytack doesn't assume PKs are integers.
        # We'll prepare this set once since we're going to query the same results in multiple ways:
        expected_pks = ["3", "2", "1"]

        results = self.sqs.all().order_by("pub_date").values("pk")
        self.assertListEqual([i["pk"] for i in results[1:11]], expected_pks)

        results = self.sqs.all().order_by("pub_date").values_list("pk")
        self.assertListEqual([i[0] for i in results[1:11]], expected_pks)

        results = self.sqs.all().order_by("pub_date").values_list("pk", flat=True)
        self.assertListEqual(results[1:11], expected_pks)

        self.assertEqual(len(connections["whoosh"].queries), 3)

    def test_manual_iter(self):
        self.sb.update(self.wmmi, self.sample_objs)
        results = self.sqs.auto_query("Indexed!")

        reset_search_queries()
        self.assertEqual(len(connections["whoosh"].queries), 0)
        results = [int(result.pk) for result in results._manual_iter()]
        self.assertEqual(sorted(results), [1, 2, 3])
        self.assertEqual(len(connections["whoosh"].queries), 1)

    def test_fill_cache(self):
        self.sb.update(self.wmmi, self.sample_objs)

        reset_search_queries()
        self.assertEqual(len(connections["whoosh"].queries), 0)
        results = self.sqs.auto_query("Indexed!")
        self.assertEqual(len(results._result_cache), 0)
        self.assertEqual(len(connections["whoosh"].queries), 0)
        results._fill_cache(0, 10)
        self.assertEqual(
            len([result for result in results._result_cache if result is not None]), 3
        )
        self.assertEqual(len(connections["whoosh"].queries), 1)
        results._fill_cache(10, 20)
        self.assertEqual(
            len([result for result in results._result_cache if result is not None]), 3
        )
        self.assertEqual(len(connections["whoosh"].queries), 2)

    def test_cache_is_full(self):
        self.sb.update(self.wmmi, self.sample_objs)

        reset_search_queries()
        self.assertEqual(len(connections["whoosh"].queries), 0)
        self.assertEqual(self.sqs._cache_is_full(), False)
        results = self.sqs.auto_query("Indexed!")
        result_list = [i for i in iter(results)]
        self.assertEqual(results._cache_is_full(), True)
        self.assertEqual(len(connections["whoosh"].queries), 1)

    def test_count(self):
        more_samples = []

        for i in range(1, 50):
            mock = MockModel()
            mock.id = i
            mock.author = "daniel%s" % i
            mock.pub_date = date(2009, 2, 25) - timedelta(days=i)
            more_samples.append(mock)

        self.sb.update(self.wmmi, more_samples)

        reset_search_queries()
        self.assertEqual(len(connections["whoosh"].queries), 0)
        results = self.sqs.all()
        self.assertEqual(len(results), 49)
        self.assertEqual(results._cache_is_full(), False)
        self.assertEqual(len(connections["whoosh"].queries), 1)

    def test_query_generation(self):
        sqs = self.sqs.filter(
            SQ(content=AutoQuery("hello world")) | SQ(title=AutoQuery("hello world"))
        )
        self.assertEqual(
            sqs.query.build_query(), "((hello world) OR title:(hello world))"
        )

    def test_result_class(self):
        self.sb.update(self.wmmi, self.sample_objs)

        # Assert that we're defaulting to ``SearchResult``.
        sqs = self.sqs.all()
        self.assertTrue(isinstance(sqs[0], SearchResult))

        # Custom class.
        sqs = self.sqs.result_class(MockSearchResult).all()
        self.assertTrue(isinstance(sqs[0], MockSearchResult))

        # Reset to default.
        sqs = self.sqs.result_class(None).all()
        self.assertTrue(isinstance(sqs[0], SearchResult))


class LiveWhooshMultiSearchQuerySetTestCase(WhooshTestCase):
    fixtures = ["bulk_data.json"]

    def setUp(self):
        super().setUp()

        # Stow.
        self.old_ui = connections["whoosh"].get_unified_index()
        self.ui = UnifiedIndex()
        self.wmmi = WhooshMockSearchIndex()
        self.wamsi = WhooshAnotherMockSearchIndex()
        self.ui.build(indexes=[self.wmmi, self.wamsi])
        self.sb = connections["whoosh"].get_backend()
        connections["whoosh"]._index = self.ui

        self.sb.setup()
        self.raw_whoosh = self.sb.index
        self.parser = QueryParser(self.sb.content_field_name, schema=self.sb.schema)
        self.sb.delete_index()

        self.wmmi.update(using="whoosh")
        self.wamsi.update(using="whoosh")

        self.sqs = SearchQuerySet("whoosh")

    def tearDown(self):
        connections["whoosh"]._index = self.old_ui
        super().tearDown()

    def test_searchquerysets_with_models(self):
        sqs = self.sqs.all()
        self.assertEqual(sqs.query.build_query(), "*")
        self.assertEqual(len(sqs), 25)

        sqs = self.sqs.models(MockModel)
        self.assertEqual(sqs.query.build_query(), "*")
        self.assertEqual(len(sqs), 23)

        sqs = self.sqs.models(AnotherMockModel)
        self.assertEqual(sqs.query.build_query(), "*")
        self.assertEqual(len(sqs), 2)


class LiveWhooshMoreLikeThisTestCase(WhooshTestCase):
    fixtures = ["bulk_data.json"]

    def setUp(self):
        super().setUp()

        # Stow.
        self.old_ui = connections["whoosh"].get_unified_index()
        self.ui = UnifiedIndex()
        self.wmmi = WhooshMockSearchIndex()
        self.wamsi = WhooshAnotherMockSearchIndex()
        self.ui.build(indexes=[self.wmmi, self.wamsi])
        self.sb = connections["whoosh"].get_backend()
        connections["whoosh"]._index = self.ui

        self.sb.setup()
        self.raw_whoosh = self.sb.index
        self.parser = QueryParser(self.sb.content_field_name, schema=self.sb.schema)
        self.sb.delete_index()

        self.wmmi.update()
        self.wamsi.update()

        self.sqs = SearchQuerySet("whoosh")

    def tearDown(self):
        connections["whoosh"]._index = self.old_ui
        super().tearDown()

    # We expect failure here because, despite not changing the code, Whoosh
    # 2.5.1 returns incorrect counts/results. Huzzah.
    @unittest.expectedFailure
    def test_more_like_this(self):
        mlt = self.sqs.more_like_this(MockModel.objects.get(pk=22))
        self.assertEqual(mlt.count(), 22)
        self.assertEqual(
            sorted([result.pk for result in mlt]),
            sorted(
                [
                    "9",
                    "8",
                    "7",
                    "6",
                    "5",
                    "4",
                    "3",
                    "2",
                    "1",
                    "21",
                    "20",
                    "19",
                    "18",
                    "17",
                    "16",
                    "15",
                    "14",
                    "13",
                    "12",
                    "11",
                    "10",
                    "23",
                ]
            ),
        )
        self.assertEqual(len([result.pk for result in mlt]), 22)

        alt_mlt = self.sqs.filter(name="daniel3").more_like_this(
            MockModel.objects.get(pk=13)
        )
        self.assertEqual(alt_mlt.count(), 8)
        self.assertEqual(
            sorted([result.pk for result in alt_mlt]),
            sorted(["4", "3", "22", "19", "17", "16", "10", "23"]),
        )
        self.assertEqual(len([result.pk for result in alt_mlt]), 8)

        alt_mlt_with_models = self.sqs.models(MockModel).more_like_this(
            MockModel.objects.get(pk=11)
        )
        self.assertEqual(alt_mlt_with_models.count(), 22)
        self.assertEqual(
            sorted([result.pk for result in alt_mlt_with_models]),
            sorted(
                [
                    "9",
                    "8",
                    "7",
                    "6",
                    "5",
                    "4",
                    "3",
                    "2",
                    "1",
                    "22",
                    "21",
                    "20",
                    "19",
                    "18",
                    "17",
                    "16",
                    "15",
                    "14",
                    "13",
                    "12",
                    "10",
                    "23",
                ]
            ),
        )
        self.assertEqual(len([result.pk for result in alt_mlt_with_models]), 22)

        if hasattr(MockModel.objects, "defer"):
            # Make sure MLT works with deferred bits.
            mi = MockModel.objects.defer("foo").get(pk=22)
            deferred = self.sqs.models(MockModel).more_like_this(mi)
            self.assertEqual(deferred.count(), 22)
            self.assertEqual(
                sorted([result.pk for result in deferred]),
                sorted(
                    [
                        "9",
                        "8",
                        "7",
                        "6",
                        "5",
                        "4",
                        "3",
                        "2",
                        "1",
                        "21",
                        "20",
                        "19",
                        "18",
                        "17",
                        "16",
                        "15",
                        "14",
                        "13",
                        "12",
                        "11",
                        "10",
                        "23",
                    ]
                ),
            )
            self.assertEqual(len([result.pk for result in deferred]), 22)

        # Ensure that swapping the ``result_class`` works.
        self.assertTrue(
            isinstance(
                self.sqs.result_class(MockSearchResult).more_like_this(
                    MockModel.objects.get(pk=21)
                )[0],
                MockSearchResult,
            )
        )


@override_settings(DEBUG=True)
class LiveWhooshAutocompleteTestCase(WhooshTestCase):
    fixtures = ["bulk_data.json"]

    def setUp(self):
        super().setUp()

        # Stow.
        self.old_ui = connections["whoosh"].get_unified_index()
        self.ui = UnifiedIndex()
        self.wacsi = WhooshAutocompleteMockModelSearchIndex()
        self.ui.build(indexes=[self.wacsi])
        self.sb = connections["whoosh"].get_backend()
        connections["whoosh"]._index = self.ui

        # Stow.
        import haystack

        self.sb.setup()
        self.sqs = SearchQuerySet("whoosh")

        # Wipe it clean.
        self.sqs.query.backend.clear()

        self.wacsi.update(using="whoosh")

    def tearDown(self):
        connections["whoosh"]._index = self.old_ui
        super().tearDown()

    def test_autocomplete(self):
        autocomplete = self.sqs.autocomplete(text_auto="mod")
        self.assertEqual(autocomplete.count(), 5)
        self.assertEqual(
            [result.pk for result in autocomplete], ["1", "12", "6", "7", "14"]
        )
        self.assertTrue("mod" in autocomplete[0].text.lower())
        self.assertTrue("mod" in autocomplete[1].text.lower())
        self.assertTrue("mod" in autocomplete[2].text.lower())
        self.assertTrue("mod" in autocomplete[3].text.lower())
        self.assertTrue("mod" in autocomplete[4].text.lower())
        self.assertEqual(len([result.pk for result in autocomplete]), 5)

    def test_edgengram_regression(self):
        autocomplete = self.sqs.autocomplete(text_auto="ngm")
        self.assertEqual(autocomplete.count(), 0)

    def test_extra_whitespace(self):
        autocomplete = self.sqs.autocomplete(text_auto="mod ")
        self.assertEqual(autocomplete.count(), 5)


class WhooshRoundTripSearchIndex(indexes.SearchIndex, indexes.Indexable):
    text = indexes.CharField(document=True, default="")
    name = indexes.CharField()
    is_active = indexes.BooleanField()
    post_count = indexes.IntegerField()
    average_rating = indexes.FloatField()
    price = indexes.DecimalField()
    pub_date = indexes.DateField()
    created = indexes.DateTimeField()
    tags = indexes.MultiValueField()
    sites = indexes.MultiValueField()
    # For a regression involving lists with nothing in them.
    empty_list = indexes.MultiValueField()

    def get_model(self):
        return MockModel

    def prepare(self, obj):
        prepped = super().prepare(obj)
        prepped.update(
            {
                "text": "This is some example text.",
                "name": "Mister Pants",
                "is_active": True,
                "post_count": 25,
                "average_rating": 3.6,
                "price": Decimal("24.99"),
                "pub_date": date(2009, 11, 21),
                "created": datetime(2009, 11, 21, 21, 31, 00),
                "tags": ["staff", "outdoor", "activist", "scientist"],
                "sites": [3, 5, 1],
                "empty_list": [],
            }
        )
        return prepped


@override_settings(DEBUG=True)
class LiveWhooshRoundTripTestCase(WhooshTestCase):
    def setUp(self):
        super().setUp()

        # Stow.
        self.old_ui = connections["whoosh"].get_unified_index()
        self.ui = UnifiedIndex()
        self.wrtsi = WhooshRoundTripSearchIndex()
        self.ui.build(indexes=[self.wrtsi])
        self.sb = connections["whoosh"].get_backend()
        connections["whoosh"]._index = self.ui

        self.sb.setup()
        self.raw_whoosh = self.sb.index
        self.parser = QueryParser(self.sb.content_field_name, schema=self.sb.schema)
        self.sb.delete_index()

        self.sqs = SearchQuerySet("whoosh")

        # Wipe it clean.
        self.sqs.query.backend.clear()

        # Fake indexing.
        mock = MockModel()
        mock.id = 1
        self.sb.update(self.wrtsi, [mock])

    def tearDown(self):
        super().tearDown()

    def test_round_trip(self):
        results = self.sqs.filter(id="core.mockmodel.1")

        # Sanity check.
        self.assertEqual(results.count(), 1)

        # Check the individual fields.
        result = results[0]
        self.assertEqual(result.id, "core.mockmodel.1")
        self.assertEqual(result.text, "This is some example text.")
        self.assertEqual(result.name, "Mister Pants")
        self.assertEqual(result.is_active, True)
        self.assertEqual(result.post_count, 25)
        self.assertEqual(result.average_rating, 3.6)
        self.assertEqual(result.price, "24.99")
        self.assertEqual(result.pub_date, datetime(2009, 11, 21, 0, 0))
        self.assertEqual(result.created, datetime(2009, 11, 21, 21, 31, 00))
        self.assertEqual(result.tags, ["staff", "outdoor", "activist", "scientist"])
        self.assertEqual(result.sites, ["3", "5", "1"])
        self.assertEqual(result.empty_list, [])

        # Check boolean filtering...
        results = self.sqs.filter(id="core.mockmodel.1", is_active=True)
        self.assertEqual(results.count(), 1)


@override_settings(DEBUG=True)
class LiveWhooshRamStorageTestCase(TestCase):
    def setUp(self):
        super().setUp()

        # Stow.
        self.old_whoosh_storage = settings.HAYSTACK_CONNECTIONS["whoosh"].get(
            "STORAGE", "file"
        )
        settings.HAYSTACK_CONNECTIONS["whoosh"]["STORAGE"] = "ram"

        self.old_ui = connections["whoosh"].get_unified_index()
        self.ui = UnifiedIndex()
        self.wrtsi = WhooshRoundTripSearchIndex()
        self.ui.build(indexes=[self.wrtsi])
        self.sb = connections["whoosh"].get_backend()
        connections["whoosh"]._index = self.ui

        # Stow.
        import haystack

        self.sb.setup()
        self.raw_whoosh = self.sb.index
        self.parser = QueryParser(self.sb.content_field_name, schema=self.sb.schema)

        self.sqs = SearchQuerySet("whoosh")

        # Wipe it clean.
        self.sqs.query.backend.clear()

        # Fake indexing.
        mock = MockModel()
        mock.id = 1
        self.sb.update(self.wrtsi, [mock])

    def tearDown(self):
        self.sqs.query.backend.clear()

        settings.HAYSTACK_CONNECTIONS["whoosh"]["STORAGE"] = self.old_whoosh_storage
        connections["whoosh"]._index = self.old_ui
        super().tearDown()

    def test_ram_storage(self):
        results = self.sqs.filter(id="core.mockmodel.1")

        # Sanity check.
        self.assertEqual(results.count(), 1)

        # Check the individual fields.
        result = results[0]
        self.assertEqual(result.id, "core.mockmodel.1")
        self.assertEqual(result.text, "This is some example text.")
        self.assertEqual(result.name, "Mister Pants")
        self.assertEqual(result.is_active, True)
        self.assertEqual(result.post_count, 25)
        self.assertEqual(result.average_rating, 3.6)
        self.assertEqual(result.pub_date, datetime(2009, 11, 21, 0, 0))
        self.assertEqual(result.created, datetime(2009, 11, 21, 21, 31, 00))
        self.assertEqual(result.tags, ["staff", "outdoor", "activist", "scientist"])
        self.assertEqual(result.sites, ["3", "5", "1"])
        self.assertEqual(result.empty_list, [])



import datetime

from django.db import models

BREED_CHOICES = [
    ("collie", "Collie"),
    ("labrador", "Labrador"),
    ("pembroke", "Pembroke Corgi"),
    ("shetland", "Shetland Sheepdog"),
    ("border", "Border Collie"),
]


class Dog(models.Model):
    breed = models.CharField(max_length=255, choices=BREED_CHOICES)
    name = models.CharField(max_length=255)
    owner_last_name = models.CharField(max_length=255, blank=True)
    birth_date = models.DateField(default=datetime.date.today)
    bio = models.TextField(blank=True)
    public = models.BooleanField(default=True)
    created = models.DateTimeField(default=datetime.datetime.now)
    updated = models.DateTimeField(default=datetime.datetime.now)

    def __str__(self):
        return self.full_name()

    @models.permalink
    def get_absolute_url(self):
        return ("dog_detail", [], {"id": self.id})

    def full_name(self):
        if self.owner_last_name:
            return "%s %s" % (self.name, self.owner_last_name)

        return self.name


class Toy(models.Model):
    dog = models.ForeignKey(Dog, on_delete=models.CASCADE, related_name="toys")
    name = models.CharField(max_length=60)

    def __str__(self):
        return "%s's %s" % (self.dog.name, self.name)



#!/usr/bin/env python
import sys
from itertools import chain
from urllib.parse import urljoin

import requests

if len(sys.argv) != 2:
    print("Usage: %s SOLR_VERSION" % sys.argv[0], file=sys.stderr)
    sys.exit(1)

solr_version = sys.argv[1]
tarball = "solr-{0}.tgz".format(solr_version)
dist_path = "lucene/solr/{0}/{1}".format(solr_version, tarball)

download_url = urljoin("https://archive.apache.org/dist/", dist_path)
mirror_response = requests.get(
    "https://www.apache.org/dyn/mirrors/mirrors.cgi/%s?asjson=1" % dist_path
)

if not mirror_response.ok:
    print(
        "Apache mirror request returned HTTP %d" % mirror_response.status_code,
        file=sys.stderr,
    )
    sys.exit(1)

mirror_data = mirror_response.json()

# Since the Apache mirrors are often unreliable and releases may disappear without notice we'll
# try the preferred mirror, all of the alternates and backups, and fall back to the main Apache
# archive server:
for base_url in chain(
    (mirror_data["preferred"],),
    mirror_data["http"],
    mirror_data["backup"],
    ("https://archive.apache.org/dist/",),
):
    test_url = urljoin(base_url, mirror_data["path_info"])

    # The Apache mirror script's response format has recently changed to exclude the actual file paths:
    if not test_url.endswith(tarball):
        test_url = urljoin(test_url, dist_path)

    try:
        if requests.head(test_url, allow_redirects=True).status_code == 200:
            download_url = test_url
            break
    except requests.exceptions.ConnectionError:
        continue
else:
    print("None of the Apache mirrors have %s" % dist_path, file=sys.stderr)
    sys.exit(1)

print(download_url)



from bare_bones_app.models import Cat

from haystack import indexes


# For the most basic usage, you can use a subclass of
# `haystack.indexes.BasicSearchIndex`, whose only requirement will be that
# you create a `search/indexes/bare_bones_app/cat_text.txt` data template
# for indexing.
class CatIndex(indexes.BasicSearchIndex, indexes.Indexable):
    def get_model(self):
        return Cat



from django.db import models


class Foo(models.Model):
    title = models.CharField(max_length=255)
    body = models.TextField()

    def __str__(self):
        return self.title


class Bar(models.Model):
    author = models.CharField(max_length=255)
    content = models.TextField()

    def __str__(self):
        return self.author



import datetime
import os
import unittest
from io import StringIO
from tempfile import mkdtemp
from unittest.mock import patch

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.core.management import call_command as real_call_command
from django.core.management.base import CommandError
from django.test import TestCase
from whoosh.qparser import QueryParser

from haystack import connections, constants, indexes
from haystack.utils.loading import UnifiedIndex

from ..core.models import MockModel
from .test_whoosh_backend import WhooshMockSearchIndex
from .testcases import WhooshTestCase


def call_command(*args, **kwargs):
    kwargs["using"] = ["whoosh"]
    print(args, kwargs)
    real_call_command(*args, **kwargs)


class ManagementCommandTestCase(WhooshTestCase):
    fixtures = ["bulk_data"]

    def setUp(self):
        super().setUp()

        self.old_ui = connections["whoosh"].get_unified_index()
        self.ui = UnifiedIndex()
        self.wmmi = WhooshMockSearchIndex()
        self.ui.build(indexes=[self.wmmi])
        self.sb = connections["whoosh"].get_backend()
        connections["whoosh"]._index = self.ui

        self.sb.setup()
        self.raw_whoosh = self.sb.index
        self.parser = QueryParser(self.sb.content_field_name, schema=self.sb.schema)
        self.sb.delete_index()

        self.sample_objs = MockModel.objects.all()

    def tearDown(self):
        connections["whoosh"]._index = self.old_ui
        super().tearDown()

    def verify_indexed_document_count(self, expected):
        with self.raw_whoosh.searcher() as searcher:
            count = searcher.doc_count()
            self.assertEqual(count, expected)

    def verify_indexed_documents(self):
        """Confirm that the documents in the search index match the database"""

        with self.raw_whoosh.searcher() as searcher:
            count = searcher.doc_count()
            self.assertEqual(count, 23)

            indexed_doc_ids = set(i["id"] for i in searcher.documents())
            expected_doc_ids = set(
                "core.mockmodel.%d" % i
                for i in MockModel.objects.values_list("pk", flat=True)
            )
            self.assertSetEqual(indexed_doc_ids, expected_doc_ids)

    def test_basic_commands(self):
        call_command("clear_index", interactive=False, verbosity=0)
        self.verify_indexed_document_count(0)

        call_command("update_index", verbosity=0)
        self.verify_indexed_documents()

        call_command("clear_index", interactive=False, verbosity=0)
        self.verify_indexed_document_count(0)

        call_command("rebuild_index", interactive=False, verbosity=0)
        self.verify_indexed_documents()

    def test_remove(self):
        call_command("clear_index", interactive=False, verbosity=0)
        self.verify_indexed_document_count(0)

        call_command("update_index", verbosity=0)
        self.verify_indexed_documents()

        # Remove several instances.
        MockModel.objects.get(pk=1).delete()
        MockModel.objects.get(pk=2).delete()
        MockModel.objects.get(pk=8).delete()
        self.verify_indexed_document_count(23)

        # Plain ``update_index`` doesn't fix it.
        call_command("update_index", verbosity=0)
        self.verify_indexed_document_count(23)

        # … but remove does:
        call_command("update_index", remove=True, verbosity=0)
        self.verify_indexed_document_count(20)

    def test_multiprocessing(self):
        call_command("clear_index", interactive=False, verbosity=0)
        self.verify_indexed_document_count(0)

        call_command("update_index", verbosity=2, workers=2, batchsize=5)
        self.verify_indexed_documents()



import datetime
from decimal import Decimal
from unittest.mock import Mock

from django.template import TemplateDoesNotExist
from django.test import TestCase

from haystack.fields import *
from test_haystack.core.models import (
    ManyToManyLeftSideModel,
    ManyToManyRightSideModel,
    MockModel,
    MockTag,
    OneToManyLeftSideModel,
    OneToManyRightSideModel,
)


class SearchFieldTestCase(TestCase):
    def test_get_iterable_objects_with_none(self):
        self.assertEqual([], SearchField.get_iterable_objects(None))

    def test_get_iterable_objects_with_single_non_iterable_object(self):
        obj = object()
        expected = [obj]

        self.assertEqual(expected, SearchField.get_iterable_objects(obj))

    def test_get_iterable_objects_with_list_stays_the_same(self):
        objects = [object(), object()]

        self.assertIs(objects, SearchField.get_iterable_objects(objects))

    def test_get_iterable_objects_with_django_manytomany_rel(self):
        left_model = ManyToManyLeftSideModel.objects.create()
        right_model_1 = ManyToManyRightSideModel.objects.create(name="Right side 1")
        right_model_2 = ManyToManyRightSideModel.objects.create()
        left_model.related_models.add(right_model_1)
        left_model.related_models.add(right_model_2)

        result = SearchField.get_iterable_objects(left_model.related_models)

        self.assertTrue(right_model_1 in result)
        self.assertTrue(right_model_2 in result)

    def test_get_iterable_objects_with_django_onetomany_rel(self):
        left_model = OneToManyLeftSideModel.objects.create()
        right_model_1 = OneToManyRightSideModel.objects.create(left_side=left_model)
        right_model_2 = OneToManyRightSideModel.objects.create(left_side=left_model)

        result = SearchField.get_iterable_objects(left_model.right_side)

        self.assertTrue(right_model_1 in result)
        self.assertTrue(right_model_2 in result)

    def test_resolve_attributes_lookup_with_field_that_points_to_none(self):
        related = Mock(spec=["none_field"], none_field=None)
        obj = Mock(spec=["related"], related=[related])

        field = SearchField(null=False)

        self.assertRaises(
            SearchFieldError,
            field.resolve_attributes_lookup,
            [obj],
            ["related", "none_field"],
        )

    def test_resolve_attributes_lookup_with_field_that_points_to_none_but_is_allowed_to_be_null(
        self,
    ):
        related = Mock(spec=["none_field"], none_field=None)
        obj = Mock(spec=["related"], related=[related])

        field = SearchField(null=True)

        self.assertEqual(
            [None], field.resolve_attributes_lookup([obj], ["related", "none_field"])
        )

    def test_resolve_attributes_lookup_with_field_that_points_to_none_but_has_default(
        self,
    ):
        related = Mock(spec=["none_field"], none_field=None)
        obj = Mock(spec=["related"], related=[related])

        field = SearchField(default="Default value")

        self.assertEqual(
            ["Default value"],
            field.resolve_attributes_lookup([obj], ["related", "none_field"]),
        )

    def test_resolve_attributes_lookup_with_deep_relationship(self):
        related_lvl_2 = Mock(spec=["value"], value=1)
        related = Mock(spec=["related"], related=[related_lvl_2, related_lvl_2])
        obj = Mock(spec=["related"], related=[related])

        field = SearchField()

        self.assertEqual(
            [1, 1],
            field.resolve_attributes_lookup([obj], ["related", "related", "value"]),
        )

    def test_resolve_attributes_lookup_with_deep_relationship_through_m2m(self):
        # obj.related2m:
        #   - related1
        #        .deep1
        #            .value = 1
        #   - related2
        #        .deep2
        #            .value = 2
        #   - related3
        #        .deep3
        #            .value = 3
        values = [1, 2, 3]
        deep1, deep2, deep3 = (Mock(spec=["value"], value=x) for x in values)
        related1, related2, related3 = (
            Mock(spec=["related"], related=x) for x in (deep1, deep2, deep3)
        )
        m2m_rel = Mock(
            spec=["__iter__"],
            __iter__=lambda self: iter([related1, related2, related3]),
        )
        obj = Mock(spec=["related_m2m"], related_m2m=m2m_rel)
        field = SearchField()
        self.assertEqual(
            values,
            field.resolve_attributes_lookup([obj], ["related_m2m", "related", "value"]),
        )

    def test_prepare_with_null_django_onetomany_rel(self):
        left_model = OneToManyLeftSideModel.objects.create()

        field = SearchField(model_attr="right_side__pk", null=True)
        result = field.prepare(left_model)

        self.assertEqual(None, result)


class CharFieldTestCase(TestCase):
    def test_init(self):
        try:
            foo = CharField(model_attr="foo")
        except:
            self.fail()

    def test_prepare(self):
        mock = MockModel()
        mock.user = "daniel"
        author = CharField(model_attr="user")

        self.assertEqual(author.prepare(mock), "daniel")

        # Do a lookup through the relation.
        mock_tag = MockTag.objects.create(name="primary")

        mock = MockModel()
        mock.tag = mock_tag
        tag_name = CharField(model_attr="tag__name")

        self.assertEqual(tag_name.prepare(mock), "primary")

        # Use the default.
        mock = MockModel()
        author = CharField(model_attr="author", default="")

        self.assertEqual(author.prepare(mock), "")

        # Simulate failed lookups.
        mock_tag = MockTag.objects.create(name="primary")

        mock = MockModel()
        mock.tag = mock_tag
        tag_slug = CharField(model_attr="tag__slug")

        self.assertRaises(SearchFieldError, tag_slug.prepare, mock)

        # Simulate failed lookups and ensure we don't get a UnicodeDecodeError
        # in the error message.
        mock_tag = MockTag.objects.create(name="básico")

        mock = MockModel()
        mock.tag = mock_tag
        tag_slug = CharField(model_attr="tag__slug")

        self.assertRaises(SearchFieldError, tag_slug.prepare, mock)

        # Simulate default='foo'.
        mock = MockModel()
        default = CharField(default="foo")

        self.assertEqual(default.prepare(mock), "foo")

        # Simulate null=True.
        mock = MockModel()
        empty = CharField(null=True)

        self.assertEqual(empty.prepare(mock), None)

        mock = MockModel()
        mock.user = None
        author = CharField(model_attr="user", null=True)

        self.assertEqual(author.prepare(mock), None)


class NgramFieldTestCase(TestCase):
    def test_init(self):
        try:
            foo = NgramField(model_attr="foo")
        except:
            self.fail()

        self.assertRaises(SearchFieldError, NgramField, faceted=True)

    def test_prepare(self):
        mock = MockModel()
        mock.user = "daniel"
        author = NgramField(model_attr="user")

        self.assertEqual(author.prepare(mock), "daniel")

        # Do a lookup through the relation.
        mock_tag = MockTag.objects.create(name="primary")

        mock = MockModel()
        mock.tag = mock_tag
        tag_name = NgramField(model_attr="tag__name")

        self.assertEqual(tag_name.prepare(mock), "primary")

        # Use the default.
        mock = MockModel()
        author = NgramField(model_attr="author", default="")

        self.assertEqual(author.prepare(mock), "")

        # Simulate failed lookups.
        mock_tag = MockTag.objects.create(name="primary")

        mock = MockModel()
        mock.tag = mock_tag
        tag_slug = NgramField(model_attr="tag__slug")

        self.assertRaises(SearchFieldError, tag_slug.prepare, mock)

        # Simulate default='foo'.
        mock = MockModel()
        default = NgramField(default="foo")

        self.assertEqual(default.prepare(mock), "foo")

        # Simulate null=True.
        mock = MockModel()
        empty = NgramField(null=True)

        self.assertEqual(empty.prepare(mock), None)

        mock = MockModel()
        mock.user = None
        author = NgramField(model_attr="user", null=True)

        self.assertEqual(author.prepare(mock), None)


class EdgeNgramFieldTestCase(TestCase):
    def test_init(self):
        try:
            foo = EdgeNgramField(model_attr="foo")
        except:
            self.fail()

        self.assertRaises(SearchFieldError, EdgeNgramField, faceted=True)

    def test_prepare(self):
        mock = MockModel()
        mock.user = "daniel"
        author = EdgeNgramField(model_attr="user")

        self.assertEqual(author.prepare(mock), "daniel")

        # Do a lookup through the relation.
        mock_tag = MockTag.objects.create(name="primary")

        mock = MockModel()
        mock.tag = mock_tag
        tag_name = EdgeNgramField(model_attr="tag__name")

        self.assertEqual(tag_name.prepare(mock), "primary")

        # Use the default.
        mock = MockModel()
        author = EdgeNgramField(model_attr="author", default="")

        self.assertEqual(author.prepare(mock), "")

        # Simulate failed lookups.
        mock_tag = MockTag.objects.create(name="primary")

        mock = MockModel()
        mock.tag = mock_tag
        tag_slug = EdgeNgramField(model_attr="tag__slug")

        self.assertRaises(SearchFieldError, tag_slug.prepare, mock)

        # Simulate default='foo'.
        mock = MockModel()
        default = EdgeNgramField(default="foo")

        self.assertEqual(default.prepare(mock), "foo")

        # Simulate null=True.
        mock = MockModel()
        empty = EdgeNgramField(null=True)

        self.assertEqual(empty.prepare(mock), None)

        mock = MockModel()
        mock.user = None
        author = EdgeNgramField(model_attr="user", null=True)

        self.assertEqual(author.prepare(mock), None)


class IntegerFieldTestCase(TestCase):
    def test_init(self):
        try:
            foo = IntegerField(model_attr="foo")
        except:
            self.fail()

    def test_prepare(self):
        mock = MockModel()
        mock.pk = 1
        pk = IntegerField(model_attr="pk")

        self.assertEqual(pk.prepare(mock), 1)

        # Simulate failed lookups.
        mock_tag = MockTag.objects.create(name="primary")

        mock = MockModel()
        mock.tag = mock_tag
        tag_count = IntegerField(model_attr="tag__count")

        self.assertRaises(SearchFieldError, tag_count.prepare, mock)

        # Simulate default=1.
        mock = MockModel()
        default = IntegerField(default=1)

        self.assertEqual(default.prepare(mock), 1)

        # Simulate null=True.
        mock = MockModel()
        pk_none = IntegerField(model_attr="pk", null=True)

        self.assertEqual(pk_none.prepare(mock), None)


class FloatFieldTestCase(TestCase):
    def test_init(self):
        try:
            foo = FloatField(model_attr="foo")
        except:
            self.fail()

    def test_prepare(self):
        mock = MockModel()
        mock.floaty = 12.5
        floaty = FloatField(model_attr="floaty")

        self.assertEqual(floaty.prepare(mock), 12.5)

        # Simulate default=1.5.
        mock = MockModel()
        default = FloatField(default=1.5)

        self.assertEqual(default.prepare(mock), 1.5)

        # Simulate null=True.
        mock = MockModel()
        floaty_none = FloatField(null=True)

        self.assertEqual(floaty_none.prepare(mock), None)


class DecimalFieldTestCase(TestCase):
    def test_init(self):
        try:
            foo = DecimalField(model_attr="foo")
        except:
            self.fail()

    def test_prepare(self):
        mock = MockModel()
        mock.floaty = Decimal("12.5")
        floaty = DecimalField(model_attr="floaty")

        self.assertEqual(floaty.prepare(mock), "12.5")

        # Simulate default=1.5.
        mock = MockModel()
        default = DecimalField(default="1.5")

        self.assertEqual(default.prepare(mock), "1.5")

        # Simulate null=True.
        mock = MockModel()
        floaty_none = DecimalField(null=True)

        self.assertEqual(floaty_none.prepare(mock), None)


class BooleanFieldTestCase(TestCase):
    def test_init(self):
        try:
            foo = BooleanField(model_attr="foo")
        except:
            self.fail()

    def test_prepare(self):
        mock = MockModel()
        mock.active = True
        is_active = BooleanField(model_attr="active")

        self.assertEqual(is_active.prepare(mock), True)

        # Simulate default=True.
        mock = MockModel()
        default = BooleanField(default=True)

        self.assertEqual(default.prepare(mock), True)

        # Simulate null=True.
        mock = MockModel()
        booly_none = BooleanField(null=True)

        self.assertEqual(booly_none.prepare(mock), None)


class DateFieldTestCase(TestCase):
    def test_init(self):
        try:
            foo = DateField(model_attr="foo")
        except:
            self.fail()

    def test_convert(self):
        pub_date = DateField()
        self.assertEqual(pub_date.convert("2016-02-16"), datetime.date(2016, 2, 16))

    def test_prepare(self):
        mock = MockModel()
        mock.pub_date = datetime.date(2009, 2, 13)
        pub_date = DateField(model_attr="pub_date")

        self.assertEqual(pub_date.prepare(mock), datetime.date(2009, 2, 13))

        # Simulate default=datetime.date(2000, 1, 1).
        mock = MockModel()
        default = DateField(default=datetime.date(2000, 1, 1))

        self.assertEqual(default.prepare(mock), datetime.date(2000, 1, 1))

    def test_prepare_from_string(self):
        mock = MockModel()
        mock.pub_date = datetime.date(2016, 2, 16)
        pub_date = DateField(model_attr="pub_date")

        self.assertEqual(pub_date.prepare(mock), datetime.date(2016, 2, 16))


class DateTimeFieldTestCase(TestCase):
    def test_init(self):
        try:
            foo = DateTimeField(model_attr="foo")
        except:
            self.fail()

    def test_convert(self):
        pub_date = DateTimeField()

        self.assertEqual(
            pub_date.convert("2016-02-16T10:02:03"),
            datetime.datetime(2016, 2, 16, 10, 2, 3),
        )

    def test_prepare(self):
        mock = MockModel()
        mock.pub_date = datetime.datetime(2009, 2, 13, 10, 1, 0)
        pub_date = DateTimeField(model_attr="pub_date")

        self.assertEqual(
            pub_date.prepare(mock), datetime.datetime(2009, 2, 13, 10, 1, 0)
        )

        # Simulate default=datetime.datetime(2009, 2, 13, 10, 01, 00).
        mock = MockModel()
        default = DateTimeField(default=datetime.datetime(2000, 1, 1, 0, 0, 0))

        self.assertEqual(default.prepare(mock), datetime.datetime(2000, 1, 1, 0, 0, 0))

    def test_prepare_from_string(self):
        mock = MockModel()
        mock.pub_date = "2016-02-16T10:01:02Z"
        pub_date = DateTimeField(model_attr="pub_date")

        self.assertEqual(
            pub_date.prepare(mock), datetime.datetime(2016, 2, 16, 10, 1, 2)
        )


class MultiValueFieldTestCase(TestCase):
    def test_init(self):
        try:
            foo = MultiValueField(model_attr="foo")
        except:
            self.fail()

        self.assertRaises(SearchFieldError, MultiValueField, use_template=True)

    def test_prepare(self):
        mock = MockModel()
        mock.sites = ["3", "4", "5"]
        sites = MultiValueField(model_attr="sites")

        self.assertEqual(sites.prepare(mock), ["3", "4", "5"])

        # Simulate default=[1].
        mock = MockModel()
        default = MultiValueField(default=[1])

        self.assertEqual(default.prepare(mock), [1])

        # Simulate null=True.
        mock = MockModel()
        multy_none = MultiValueField(null=True)

        self.assertEqual(multy_none.prepare(mock), None)

    def test_convert_with_single_string(self):
        field = MultiValueField()

        self.assertEqual(["String"], field.convert("String"))

    def test_convert_with_single_int(self):
        field = MultiValueField()

        self.assertEqual([1], field.convert(1))

    def test_convert_with_list_of_strings(self):
        field = MultiValueField()

        self.assertEqual(
            ["String 1", "String 2"], field.convert(["String 1", "String 2"])
        )

    def test_convert_with_list_of_ints(self):
        field = MultiValueField()

        self.assertEqual([1, 2, 3], field.convert([1, 2, 3]))


class CharFieldWithTemplateTestCase(TestCase):
    def test_init(self):
        try:
            foo = CharField(use_template=True)
        except:
            self.fail()

        try:
            foo = CharField(use_template=True, template_name="foo.txt")
        except:
            self.fail()

        foo = CharField(use_template=True, template_name="foo.txt")
        self.assertEqual(foo.template_name, "foo.txt")

        # Test the select_template usage.
        foo = CharField(use_template=True, template_name=["bar.txt", "foo.txt"])
        self.assertEqual(foo.template_name, ["bar.txt", "foo.txt"])

    def test_prepare(self):
        mock = MockModel()
        mock.pk = 1
        mock.user = "daniel"
        template1 = CharField(use_template=True)

        self.assertRaises(SearchFieldError, template1.prepare, mock)

        template2 = CharField(use_template=True)
        template2.instance_name = "template_x"
        self.assertRaises(TemplateDoesNotExist, template2.prepare, mock)

        template3 = CharField(use_template=True)
        template3.instance_name = "template"
        self.assertEqual(template3.prepare(mock), "Indexed!\n1\n")

        template4 = CharField(use_template=True, template_name="search/indexes/foo.txt")
        template4.instance_name = "template"
        self.assertEqual(template4.prepare(mock), "FOO!\n")

        template5 = CharField(
            use_template=True, template_name=["foo.txt", "search/indexes/bar.txt"]
        )
        template5.instance_name = "template"
        self.assertEqual(template5.prepare(mock), "BAR!\n")


##############################################################################
# The following tests look like they don't do much, but it's important because
# we need to verify that the faceted variants behave like the field they
# emulate. The old-broke behavior was convert everything to string.
##############################################################################


class FacetFieldTestCase(TestCase):
    def test_init(self):
        # You shouldn't use the FacetField itself.
        try:
            foo = FacetField(model_attr="foo")
            self.fail()
        except:
            pass

        try:
            foo_exact = FacetField(facet_for="bar")
            self.fail()
        except:
            pass


class FacetCharFieldTestCase(TestCase):
    def test_init(self):
        try:
            foo = FacetCharField(model_attr="foo")
            foo_exact = FacetCharField(facet_for="bar")
        except:
            self.fail()

        self.assertEqual(foo.facet_for, None)
        self.assertEqual(foo_exact.null, True)
        self.assertEqual(foo_exact.facet_for, "bar")

    def test_prepare(self):
        mock = MockModel()
        mock.user = "daniel"
        author = FacetCharField(model_attr="user")

        self.assertEqual(author.prepare(mock), "daniel")


class FacetIntegerFieldTestCase(TestCase):
    def test_init(self):
        try:
            foo = FacetIntegerField(model_attr="foo")
            foo_exact = FacetIntegerField(facet_for="bar")
        except:
            self.fail()

        self.assertEqual(foo.facet_for, None)
        self.assertEqual(foo_exact.null, True)
        self.assertEqual(foo_exact.facet_for, "bar")

    def test_prepare(self):
        mock = MockModel()
        mock.user = "daniel"
        mock.view_count = 13
        view_count = FacetIntegerField(model_attr="view_count")

        self.assertEqual(view_count.prepare(mock), 13)


class FacetFloatFieldTestCase(TestCase):
    def test_init(self):
        try:
            foo = FacetFloatField(model_attr="foo")
            foo_exact = FacetFloatField(facet_for="bar")
        except:
            self.fail()

        self.assertEqual(foo.facet_for, None)
        self.assertEqual(foo_exact.null, True)
        self.assertEqual(foo_exact.facet_for, "bar")

    def test_prepare(self):
        mock = MockModel()
        mock.user = "daniel"
        mock.price = 25.65
        price = FacetFloatField(model_attr="price")

        self.assertEqual(price.prepare(mock), 25.65)


class FacetBooleanFieldTestCase(TestCase):
    def test_init(self):
        try:
            foo = FacetBooleanField(model_attr="foo")
            foo_exact = FacetBooleanField(facet_for="bar")
        except:
            self.fail()

        self.assertEqual(foo.facet_for, None)
        self.assertEqual(foo_exact.null, True)
        self.assertEqual(foo_exact.facet_for, "bar")

    def test_prepare(self):
        mock = MockModel()
        mock.user = "daniel"
        mock.is_active = True
        is_active = FacetBooleanField(model_attr="is_active")

        self.assertEqual(is_active.prepare(mock), True)


class FacetDateFieldTestCase(TestCase):
    def test_init(self):
        try:
            foo = FacetDateField(model_attr="foo")
            foo_exact = FacetDateField(facet_for="bar")
        except:
            self.fail()

        self.assertEqual(foo.facet_for, None)
        self.assertEqual(foo_exact.null, True)
        self.assertEqual(foo_exact.facet_for, "bar")

    def test_prepare(self):
        mock = MockModel()
        mock.user = "daniel"
        mock.created = datetime.date(2010, 10, 30)
        created = FacetDateField(model_attr="created")

        self.assertEqual(created.prepare(mock), datetime.date(2010, 10, 30))


class FacetDateTimeFieldTestCase(TestCase):
    def test_init(self):
        try:
            foo = FacetDateTimeField(model_attr="foo")
            foo_exact = FacetDateTimeField(facet_for="bar")
        except:
            self.fail()

        self.assertEqual(foo.facet_for, None)
        self.assertEqual(foo_exact.null, True)
        self.assertEqual(foo_exact.facet_for, "bar")

    def test_prepare(self):
        mock = MockModel()
        mock.user = "daniel"
        mock.created = datetime.datetime(2010, 10, 30, 3, 14, 25)
        created = FacetDateTimeField(model_attr="created")

        self.assertEqual(
            created.prepare(mock), datetime.datetime(2010, 10, 30, 3, 14, 25)
        )


class FacetMultiValueFieldTestCase(TestCase):
    def test_init(self):
        try:
            foo = FacetMultiValueField(model_attr="foo")
            foo_exact = FacetMultiValueField(facet_for="bar")
        except:
            self.fail()

        self.assertEqual(foo.facet_for, None)
        self.assertEqual(foo_exact.null, True)
        self.assertEqual(foo_exact.facet_for, "bar")

    def test_prepare(self):
        mock = MockModel()
        mock.user = "daniel"
        mock.sites = [1, 3, 4]
        sites = FacetMultiValueField(model_attr="sites")

        self.assertEqual(sites.prepare(mock), [1, 3, 4])



""""""

import unittest

import django
from django.test import TestCase

import haystack


class AppConfigCompatibilityTestCase(TestCase):
    @unittest.skipIf(
        django.VERSION >= (3, 2), "default_app_config is deprecated since django 3.2."
    )
    def testDefaultAppConfigIsDefined_whenDjangoVersionIsLessThan3_2(self):
        has_default_appconfig_attr = hasattr(haystack, "default_app_config")
        self.assertTrue(has_default_appconfig_attr)

    @unittest.skipIf(
        django.VERSION < (3, 2),
        "default_app_config should be used in versions prior to django 3.2.",
    )
    def testDefaultAppConfigIsDefined_whenDjangoVersionIsMoreThan3_2(self):
        has_default_appconfig_attr = hasattr(haystack, "default_app_config")
        self.assertFalse(has_default_appconfig_attr)
